import warnings
with warnings.catch_warnings():
    try:
        __import__('pkg_resources').declare_namespace(__name__)
    except ImportError:
        import pkgutil
        __path__ = pkgutil.extend_path(__path__, __name__)
